<?php
/**
 * PageQuery Plugin: search for and list pages, sorted by name, date, creator, etc
 *
 * @license    GPL 2 (http://www.gnu.org/licenses/gpl.html)
 * @author	   Symon Bent <hendrybadao@gmail.com>
 */
// must be run within Dokuwiki
if(!defined('DOKU_INC')) die();

if (!defined('DOKU_LF')) define('DOKU_LF', "\n");
if (!defined('DOKU_TAB')) define('DOKU_TAB', "\t");
if (!defined('DOKU_PLUGIN')) define('DOKU_PLUGIN',DOKU_INC.'lib/plugins/');

require_once(DOKU_PLUGIN.'syntax.php');
require_once(DOKU_INC.'inc/fulltext.php');

class syntax_plugin_pagequery extends DokuWiki_Syntax_Plugin {
    private $dformat = '';
    private $inwords = false;
    private $nostart = false;

	function getType() {
		return 'substition';
	}

	function getPType() {
		return 'block';
	}

	function getSort() {
		return 98;
	}

	function connectTo($mode) {
		//Default is a word cloud style, looks good when combined with plugin:box
		$this->Lexer->addSpecialPattern('\{\{pagequery>.*?\}\}', $mode, 'plugin_pagequery');
	}
    /**
     * Parses all the pagequery options:
     * Insert the pagequery markup wherever you want your list to appear. E.g:
     *
     *   {{pagequery>}}
     *
     *   {{pagequery>[query];fulltext;sort=key:direction,key2:direction;group;limit=??;cols=?;inwords;proper}}
     *
     * Parameters as follows:
     * 1. query:    any expression directly after the >; can use all Dokuwiki search options (see manual)
     * 2. fulltext: use a full-text search, instead of page_id only [default]
     * 3. sort:     keys to sort by, in order of sorting. Each key can be followed by prefered sorting order
     *              available keys:
     *                  a, ab, abc          by 1st letter, 2 letters, or 3 letters
     *                  name                by page name (no namespace) [not grouped]
     *                  page|id             by full page id, including namespace [not grouped]
     *                  ns                  by namespace (without page name)
     *                  mdate, cdate        by modified|created dates (full) [not grouped]
     *                  m[year][month][day] by modified [year][month][day]; any combination accepted
     *                  c[year][month][day] by created [year][month][day]; any combination accepted
     *                  creator             by page author
     *              date sort default to descending, string sorts to ascending
     * 4. group:    show group headers for each change in sort keys
     *              Note: keys with no duplicated cannot be grouped (i.e. name, page|id, mdate, cdate)
     * 5. limit:    maximum number of results to return
     * 6. inwords:  use real month and day names instead of just numbers
     * 7. cols:     number of columns in displayed list (max = 4)
     * 8. proper:   display page names and namespace in Proper Case (i.e. no _'s and Capitalised)
                    header/hdr = group headers only, name = page name only, both = both!
     * 9. border:   turn on borders
     *10. fullregex:only useful on page name searches; allows a raw regex mode on the full page idate
     *11. nostart:  ignore any 'start' pages in namespace
     *
     * All options are optional, and the list will default to a boring long 1-column list...
     */
	function handle($match, $state, $pos, &$handler) {
        global $conf;
        $this->dformat = $conf['dformat'];

		$match = substr($match, 12, -2); // strip markup incl. pagequery>
		$args = explode(';', $match);
        $query = $args[0];
        $fulltext = false;
        $sort = array();
        $group = false;
        $limit = 0; //Default max items displayed. '0' => no limit
        $proper = 'none';
        $cols = 1;
        $border = false;
        $title = false; // use title instead of name
        $abstract = 'tooltip';

        foreach ($args as $arg) {
            list($key, $value) = explode('=', $arg);
            switch ($key) {
                case 'fulltext':
                    $fulltext = true;
                    break;
                case 'sort':
                    $values = explode(',', $value);
                    foreach ($values as $value) {
                        list($key, $dir) = explode(':', $value);
                        $sort[] = array($key, $dir);
                    }
                    break;
                case 'group':
                    $group = true;
                    break;
                case 'limit':
                    $limit = $value;
                    break;
                case 'inwords':
                    $inwords = true;
                    break;
                case 'proper':
                    switch ($value) {
                        case 'hdr':
                        case 'header':
                        case 'group':
                            $proper = 'header';
                            break;
                        case 'name':
                        case 'page':
                            $proper = 'name';
                            break;
                        default:
                            $proper = 'both';
                    }
                    break;
                case 'cols':
                    $cols = ($value > 4) ? 4 : $value;
                    break;
                case 'border':
                    $border = true;
                    break;
				case 'fullregex':
					$fullregex = true;
					break;
                case 'nostart':
                    $nostart = true;
                    break;
                case 'title':
                    $title = true;
                    break;
                case 'abstract':
                    switch ($value) {
                        case '':
                        case 'tooltip':
                            $abstract = 'tooltip';
                            break;
                        default:
                            if ($value >= 0) {
                                $abstract = $value;
                            }
                }
            }
        }
		return array($query, $fulltext, $sort, $group, $limit, $proper,
                      $cols, $inwords, $border, $fullregex, $nostart, $title, $abstract);
	}

    function render($mode, &$renderer, $data) {
        $sortkeys = array();
        list($query, $fulltext, $sort, $group, $limit, $proper,
              $cols, $inwords, $border, $fullregex, $nostart, $title, $abstract) = $data;
        $this->inwords = $inwords;
        $this->nostart = $nostart;

        if ($mode == 'xhtml') {
            // first get a raw list of matching results
            if ($fulltext) {
                $results = array_keys(ft_pageSearch($query, $highlight));
            } else {
                // allow for raw regex mode, for power users, this searches the full page id
                if ($fullregex) {
                    $full_query = $query;
                    $pageonly = false;
                } else {
                    list($full_query, $pageonly) = $this->_parse_ns_query($query);
                }
                $results = $this->_pageLookup($full_query, $pageonly);
            }
            // basic result page markup (always needed)
            $border_css = ($border === false) ? ' style="border-style:none;"' : '';
            $renderer->doc .= '<div id="pagequery">' . DOKU_LF;
            $renderer->doc .= '<table><tbody><tr>' . DOKU_LF;

            // display a "no results" message if empty
            if (empty($results)) {
                $renderer->doc .= '<td><p>' . $this->getLang("no_results") .
                                  '"<strong>' . $query . '</strong>"</p></td>';
            } else {
                // prepare the necessary sorting arrays, as per options
                list($sort_opts, $dates) = $this->_get_sort_opts($results, $sort, $title);
                // now do the sorting
                $this->_sort_results($results, $sort_opts, $dates);
                // limit the result list length if required
                // remember: this can only be done after sorting!
                $results = ($limit > 0) ? array_slice($results, 0, $limit) : $results;
                // and finally the grouping
                if ($group) {
                    $results = $this->_group_results($results, $sort_opts, $dates);
                }
                $was_heading = false;
                $idx = 0;
                $col = 0;
                $rows = ceil(count($results) / $cols);  // proposed column row size
                $width = floor(100 / $cols);
                // remove indenting if not grouped
                $indent_css = ($group === false) ? ' class="nogroup"' : '';
                $link_count = 0;    // needed by the abstract snippet section

                foreach ($results as $id) {
                    // are we ready to start a new column? (up to max allowed)
                    if ($col < $cols && $idx % $rows == 0) {
                        $ready_for_col = true;
                        $col++;
                    }
                    $is_heading = ($id[1] == '|');
                    if ($is_heading) {
                        list($hl, $heading) = explode('|', $id);
                        if ($proper == 'header' || $proper == 'both') $heading = $this->_proper($heading);
                    }

                    // Begin new column if: we are at the start, at a heading or if there is no grouping
                    if ($ready_for_col && !$was_heading) {
                        // close the previous column if necessary
                        $cont_close = (!$is_heading) ? '<strong>⇒ ... </strong>' : '';
                        $cont_open = ($idx > 0 && !$is_heading) ? "<h$hl>$heading ...</h$hl>" : '';
                        $close = ($idx > 0) ? $cont_close . '</ul></td>' . DOKU_LF : '';
                        $renderer->doc .= $close . '<td' . $border_css . ' valign="top" width="' . $width . '%">' . $cont_open . DOKU_LF;
                        $ready_for_col = false;
                        $was_heading = true;    //needed to begin a new page id list
                    }
                    if ($is_heading) {
                        // close previous sub list if necessary
                        if (!$was_heading) $renderer->doc .= '</ul>' . DOKU_LF;
                        $renderer->doc .= "<h$hl>$heading</h$hl>" . DOKU_LF;
                        $was_heading = true;
                    } else {
                        // open a new sub list if necessary
                        if ($was_heading || $idx == 0) $renderer->doc .= '<ul' . $indent_css . '>';

                        // deal with normal page links
                        $link = $this->_render_wikilink($id, $proper, $title, $abstract, $link_count);
                        $link_count++;
                        $renderer->doc .= $link;
                        $was_heading = false;
                    }
                    $idx++;
                }
                $renderer->doc .= '</ul></td>';
            }
            $renderer->doc .= '</tr>' . DOKU_LF;
            $renderer->doc .= '</tbody></table>';
            $renderer->doc .= '</div>' . DOKU_LF;
        } else {
            return false;
        }
    }
    /**
     * Renders the page link, plus tooltip, abstract, casing, etc...
     * @param string $id
     * @param bool  $proper
     * @param bool  $title
     * @param mixed $abstract
     */
    private function _render_wikilink($id, $proper, $title, $abstract, $link_count) {

        $id = (strpos($id, ':') === false) ? ':' . $id : $id;   // : needed for root pages

        // does the user want to see the "title" instead "pagename"
        if ($title) {
            $value = p_get_metadata($id, 'title', true);
            $name = (empty($value)) ? $this->_proper(noNS($id)) : $value;
        } elseif ($proper == 'name' || $proper == 'both') {
            $name = $this->_proper(noNS($id));
        } else {
            $name = '';
        }

        // show the "abstract" as a tooltip
        // -or- as a nice quoted snippet (à la blockquote plugin)
        $detail = '';
        if ($abstract == 'none') {
            $link = html_wikilink($id, $name);
        } elseif ($abstract == 'tooltip' || ($abstract > 0 && $link_count >= $abstract)) {
            $link = $this->_html_wikilink_tooltip($id, $name);
        } elseif ($abstract == 'all' || $abstract == 0 || ($abstract > 0 && $link_count < $abstract)) {
            $link = html_wikilink($id, $name);
            $detail = str_replace("\n\n", '<br/>', $this->_abstract($id));
            $more = html_wikilink($id, 'more...');
            $detail = '<blockquote>' . $detail . $more . '</blockquote>' . DOKU_LF;
        }
        return '<li>' . $link . '</li>' . DOKU_LF . $detail;
    }

    private function _proper($id) {
         $id = str_replace(':', ': ', $id);
         $id = str_replace('_', ' ', $id);
         $id = ucwords($id);
         $id = str_replace(': ', ':', $id);
         return $id;
    }
    /**
     * swap normal link title (popup) for a more useful preview
     *
     * @param string $id    page id
     * @param string $name  display name
     * @return complete href link
     */
    private function _html_wikilink_tooltip($id, $name) {
        $link = html_wikilink($id, $name);
        $tooltip = $this->_abstract($id);
        $tooltip = str_replace("\n", '  ', $tooltip);
        $link = preg_replace('/title=\".+?\"/', 'title="' . $tooltip . '"', $link, 1);
        return $link;
    }

    private function _abstract($id) {
        $abstract = p_get_metadata($id, 'description abstract', true);
        $abstract = htmlspecialchars($abstract, ENT_QUOTES, 'UTF-8');
        return $abstract;
    }

    /**
     * Parse out the namespace, and convert to a regex for array search
     * -or- ensures that normal search are only within pagename not its namespace
     *
     * @param string $query user page query
     * @return processed query with regex markup as needed for namespace recognition
     */
    private function _parse_ns_query($query) {
        $incl_ns = preg_match('`(@|ns:)(\w+(\:\w+)*)(\s.+)*`', $query, $matches);
        if ($incl_ns > 0) {
            $ns = $matches[2];
            $page = trim($matches[4]);
            if ($page == '*') $page = '.*';
            $ns_query = $ns . ":.*" . $page;
            return array($ns_query, false);
        }
        $excl_ns = preg_match('`(\^|-ns:)(\w+(\:\w+)*)(\s.+)*`', $query, $matches);
        if ($excl_ns > 0) {
            $ns = $matches[2];
            $page = trim($matches[4]);
            if ($page == '*') $page = '.*';
            $ns_query = '^(?!' . $ns . '):.*' . $page;
            return array($ns_query, false);
        }
		if ($query == '*') $query= '.*';
        return array($query, true);	// true => turns on pageonly lookup
    }

    /**
     * Builds a separate sorting array for each sortby key
     *
     * @param string    $id         wiki page id
     * @param array     $sortby     keys to use for sorting
     * @param arrayref  $sortkeys   ref to sorting arrays
     * @return void
     */
    private function _get_sort_opts($ids, $sortkeys, $title) {
        global $conf;
        $sort_opts = array();
        $dformat = array();
        $dates = array();

        $cnt = 0;
        foreach ($ids as $id) {
            $meta = p_get_metadata ($id, false, true);
            if (!isset($meta['date']['modified'])) {
                $meta['date']['modified'] = $meta['date']['created'];
            }
            if ($title && !isset($meta['title'])) {
                $meta['title'] = $this->_proper(noNS($id));
            }
            $idx = 0;
            foreach ($sortkeys as $sortkey) {
                $key = $sortkey[0];
                switch ($key) {
                    case 'a':
                        $value = $this->_first(noNS($id), 1);
                        break;
                    case 'ab':
                        $value = $this->_first(noNS($id), 2);
                        break;
                    case 'abc':
                        $value = $this->_first(noNS($id), 3);
                        break;
                    case 'name':
                        $value = ($title) ? $meta['title'] : noNS($id);
                        break;
                    case 'id':
                    case 'page':
                        $value = $id;
                        break;
                    case 'ns':
                        $value = getNS($id);
                        if (empty($value)) $value = '(' . $conf['start'] . ')';
                        break;
                    case 'creator':
                        $value = $meta['creator'];
                        break;
                    case 'mdate':
                        $value = $meta['date']['modified'];
                        break;
                    case 'cdate':
                        $value = $meta['date']['created'];
                        break;
                    default:
                        $dtype = $key[0];
                        if ($dtype == 'c' || $dtype == 'm') {
                            // only set date once per id (not per sort column--date remains same across columns)
                            if ( ! isset($dates[$cnt])) {
                                if ($dtype == 'c') {
                                    $date = $meta['date']['created'];
                                } else {
                                    $date = $meta['date']['modified'];
                                }
                                $dates[$cnt] = $date;
                            } else {
                                $date = $dates[$cnt];
                            }
                            // the date formats are cached on first use; allows one of each key type per set
                            if ( ! isset($dformat[$key])) {
                                $dformat[$key] = $this->_date_format($key, $this->inwords);
                                $sort_opts[$idx]['inwords'] = $dformat[$key][1];
                            }
                            $value = strftime($dformat[$key][0], $date);
                        }
                }
                $sort_opts[$idx]['keys'][] = $value;
                $idx++;
            }
            $cnt++;
        }
        $idx = 0;
        foreach ($sortkeys as $sortkey) {
            list($key, $dir) = $sortkey;
            // now the sort direction
            switch ($dir) {
                case 'a':
                case 'asc':
                    $dir = SORT_ASC;
                    break;
                case 'd':
                case 'desc':
                    $dir = SORT_DESC;
                    break;
                default:
                    // sort descending by default; text ascending
                    // watch for other options beginning with c or m ....!
                    if ($key[0] == 'c' || $key[0] == 'm') {
                        $dir = SORT_DESC;
                    } else {
                        $dir = SORT_ASC;
                    }
            }
            $sort_opts[$idx]['dir'] = $dir;
            // and finally set the sort array's data type (for correct markup purposes)
            switch ($key) {
                // full-text and full dates cannot meaningfully be grouped (no duplicates!)
                case 'mdate':
                case 'cdate':
                case 'name':
                case 'id':
                case 'page':
                    $group = false;
                    break;
                case 'ns':
                    $group = 'ns';
                    break;
                default:
                    $group = true;
            }
            $sort_opts[$idx]['group'] = $group;
            // and now the sort data type
            $is_ns = false;
            switch ($key) {
                case 'mdate':
                case 'cdate':
                    $type = SORT_NUMERIC;
                    break;
                default:
                    $type = SORT_STRING;
            }
            $sort_opts[$idx]['type'] = $type;
            $idx++;
        }
        return array($sort_opts, $dates);
    }

    // returns first $count letters from $text
    private function _first($text, $count) {
        if ($count > 0) {
            return substr($text, 0, $count);
        }
    }

    // parse the c|m-year-month-day option
    private function _date_format($key, $inwords = false) {
        $year = (strpos($key, 'year') !== false) ? '%Y' : '';
        $month = (strpos($key, 'month') !== false) ? '%m' : '';
        $day = (strpos($key, 'day') !== false) ? '%d' : '';
        $format = $year . (($year != '' && ($month != '' || $day != '')) ? '-' : '');
        $format .= $month . ($month != '' && $day != '' ? '-' : '');
        $format .= $day;
        // provide month and day format in real words if required
        // used for display only, the $key is used for sorting!
        $wformat = '';
        if ($inwords) {
            switch ($format) {
                case '%m':
                    $wformat = "%B";
                    break;
                case '%d':
                    $wformat = "%#d, %A ";
                    break;
                case '%Y-%m':
                    $wformat = "%B %Y";
                    break;
                case '%m-%d':
                    $wformat= "%B %#d, %A ";
                    break;
                case '%Y-%m-%d':
                    $wformat = "%A, %B %#d, %Y";
                    break;
            }
        }
        return array($format, $wformat);
    }

    // sort by each sorting array, in order
    private function _sort_results(&$ids, &$sort_opts, &$dates) {
        if ( ! empty($sort_opts)) {
            $args = array();
            foreach ($sort_opts as &$sort_opt) {
                $args[] = &$sort_opt['keys'];
                $args[] = &$sort_opt['dir'];
                $args[] = &$sort_opt['type'];
            }
            $args[] = &$dates;
            $args[] = &$ids;
            call_user_func_array('array_multisort', $args);
        }
    }

    private function _group_results(&$ids, &$sort_opts, &$dates) {
        $prevs = array();
        $results = array();
        $level = count($sort_opts) - 1;
        $idx = 0;
        foreach($ids as $id) {
            $this->_add_heading($results, $sort_opts, $dates, $level, $idx, $prevs);
            $results[] = $id;
            $idx++;
        }
        return $results;
    }

    private function _add_heading(&$results, &$sort_opts, &$dates, $level, $idx, &$prevs) {
        if ($level > 0) {
            // recurse to find all parent headings
            $this->_add_heading($results, $sort_opts, $dates, $level - 1, $idx, $prevs);
        }
        $group = $sort_opts[$level]['group'];
        // only if this sort array can be grouped....
        if ($group === true) {
            $prev = (isset($prevs[$level])) ? $prevs[$level] : '';
            $cur = $sort_opts[$level]['keys'][$idx];
            if ($cur != $prev) {
                $prevs[$level] = $cur;
                $hl = ($level + 1);
                $inwords = $sort_opts[$level]['inwords'];
                if ($this->inwords && ! empty($inwords)) $cur = strftime($inwords, $dates[$idx]);
                $results[] = $hl . '|' . $cur;
            }
        } elseif ($group == 'ns') {
            $prev = (isset($prevs[$level])) ? $prevs[$level] : '';
            $cur = $sort_opts[$level]['keys'][$idx];
            if ($cur != $prev) {
                $prevs[$level] = $cur;
                $cur_ns = explode(':', $cur);
                $prev_ns = explode(':', $prev);
                // we only show namespaces that are different from the previous heading
                for ($i= 0; $i < count($cur_ns); $i++) {
                    if ($cur_ns[$i] != $prev_ns[$i]) {
                        $hl = $level + $i + 1;
                        $hl = ($hl > 6) ? 6 : $hl;
                        $results[] = $hl . '|' . $cur_ns[$i];
                    }
                }
            }
        }
    }


    /**
     * heavily customised version of _ft_pageLookup in inc/fulltext.php
     * I just removed the preg_quote line to enable regexes in search query...
     * and the unnecessary usort line as I have my own sort routines
     *
     */
    private function _pageLookup($query, $pageonly) {
        global $conf;

        $query = trim($query);
        $pages = file($conf['indexdir'].'/page.idx');
        $pages = array_values(preg_grep('/'.$query.'/', $pages));

        $cnt = count($pages);
        for($i = 0; $i < $cnt; $i++){
            if($pageonly){
                if(!preg_match('/'.$query.'/',noNS($pages[$i]))){
                    unset($pages[$i]);
                    continue;
                }
            }
            if(!page_exists($pages[$i])){
                unset($pages[$i]);
                continue;
            }
        }

        $pages = array_filter($pages, 'isVisiblePage'); // discard hidden pages
        if(!count($pages)) return array();

        $pages = array_map('trim',$pages);

        // check ACL permissions and remove any 'start' pages if req'd
        $start = $conf['start'];
        $pos = strlen($start);
        foreach($pages as $idx => $name) {
            if ($this->nostart && substr($name, -$pos) == $start) {
                unset($pages[$idx]);
            } elseif(auth_quickaclcheck(trim($pages[$idx])) < AUTH_READ){
                unset($pages[$idx]);
            }
        }
        return $pages;
    }
}
